# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, cast

from dunia.error import PlaywrightTimeoutError
from dunia.page import AsyncEventContextManager, Download, FileChooser, Frame

if TYPE_CHECKING:
    from typing import Any, Literal

    import playwright.async_api as playwright
    from typing_extensions import Self

    from dunia.element import Node


@dataclass(slots=True, frozen=True)
class PlaywrightPage:
    page: playwright.Page

    async def query_selector(self, selector: str) -> Node | None:
        if handle := await self.page.query_selector(selector):
            return PlaywrightElementHandle(handle)

        return None

    async def query_selector_all(self, selector: str) -> list[Node]:
        return [
            PlaywrightElementHandle(handle)
            for handle in await self.page.query_selector_all(selector)
        ]

    async def text_content(
        self, selector: str, *, timeout: float | None = None
    ) -> str | None:
        try:
            text = await self.page.text_content(selector, timeout=timeout)  # type: ignore
        except PlaywrightTimeoutError:
            return None
        else:
            return text

    async def inner_text(
        self, selector: str, *, timeout: float | None = None
    ) -> str | None:
        try:
            text = await self.page.inner_text(selector, timeout=timeout)  # type: ignore
        except PlaywrightTimeoutError:
            return None
        else:
            return text

    async def get_attribute(
        self, selector: str, name: str, *, timeout: float | None = None
    ) -> str | None:
        try:
            attribute = await self.page.get_attribute(selector, name, timeout=timeout)  # type: ignore
        except PlaywrightTimeoutError:
            return None
        else:
            return attribute

    async def click(self, selector: str, *, timeout: float | None = None) -> None:
        await self.page.click(selector, timeout=timeout)  # type: ignore

    async def press(
        self, selector: str, key: str, *, timeout: float | None = None
    ) -> None:
        await self.page.press(selector, key, timeout=timeout)  # type: ignore

    async def fill(
        self, selector: str, value: str, *, timeout: float | None = None
    ) -> None:
        await self.page.fill(selector, value, timeout=timeout)  # type: ignore

    async def evaluate(self, expression: str) -> Any:
        return await self.page.evaluate(expression)

    async def type(
        self, selector: str, text: str, *, timeout: float | None = None
    ) -> None:
        await self.page.type(selector, text, timeout=timeout)  # type: ignore

    async def inner_html(
        self, selector: str, *, timeout: float | None = None
    ) -> str | None:
        try:
            html = await self.page.inner_html(selector, timeout=timeout)  # type: ignore
        except PlaywrightTimeoutError:
            return None
        else:
            return html

    @property
    def mouse(self) -> playwright.Mouse:
        return self.page.mouse

    async def content(self) -> str:
        return await self.page.content()

    async def set_content(
        self,
        html: str,
        *,
        timeout: float | None = None,
        wait_until: Literal[
            "commit", "domcontentloaded", "load", "networkidle"
        ] = "load",
    ) -> None:
        await self.page.set_content(html, timeout=timeout, wait_until=wait_until)  # type: ignore

    async def goto(
        self,
        url: str,
        *,
        timeout: float | None = None,
        wait_until: Literal[
            "commit", "domcontentloaded", "load", "networkidle"
        ] = "load",
    ) -> None:
        await self.page.goto(url, timeout=timeout, wait_until=wait_until)  # type: ignore

    async def close(self) -> None:
        await self.page.close()

    async def pause(self) -> None:
        await self.page.pause()

    @property
    def url(self) -> str:
        return self.page.url

    async def screenshot(
        self,
        path: str,
        *,
        full_page: bool = False,
        timeout: float | None = None,
    ) -> bytes | None:
        try:
            screenshot = await self.page.screenshot(
                path=path, full_page=full_page, timeout=timeout  # type: ignore
            )
        except PlaywrightTimeoutError:
            return None

        return screenshot

    async def wait_for_load_state(
        self,
        state: Literal["domcontentloaded", "load", "networkidle"],
        *,
        timeout: float | None = None,
    ) -> None:
        await self.page.wait_for_load_state(state, timeout=timeout)  # type: ignore

    async def wait_for_selector(
        self,
        selector: str,
        *,
        timeout: float | None = None,
        state: Literal["attached", "detached", "hidden", "visible"] | None = None,
    ) -> Node | None:
        if handle := await self.page.wait_for_selector(
            selector, timeout=timeout, state=state  # type: ignore
        ):
            return PlaywrightElementHandle(handle)

        return None

    async def wait_for_timeout(self, timeout: float) -> None:
        await self.page.wait_for_timeout(timeout=timeout)

    async def select_option(
        self,
        selector: str,
        value: str | list[str] | None = None,
        *,
        label: str | list[str] | None = None,
        timeout: float | None = None,
    ) -> list[str]:
        return await self.page.select_option(
            selector, value=value, label=label, timeout=timeout  # type: ignore
        )

    def frame(self, name: str) -> Frame | None:
        return cast(Frame | None, self.page.frame(name))

    def expect_navigation(
        self, *, timeout: float | None = None
    ) -> AsyncEventContextManager[None]:
        return cast(
            AsyncEventContextManager[None],
            self.page.expect_navigation(timeout=timeout),  # type: ignore
        )

    def expect_request_finished(
        self, *, timeout: float | None = None
    ) -> AsyncEventContextManager[None]:
        return cast(
            AsyncEventContextManager[None],
            self.page.expect_request_finished(timeout=timeout),  # type: ignore
        )

    def expect_download(
        self, *, timeout: float | None = None
    ) -> AsyncEventContextManager[Download]:
        return cast(
            AsyncEventContextManager[Download],
            self.page.expect_download(timeout=timeout),  # type: ignore
        )

    def expect_file_chooser(
        self, *, timeout: float | None = None
    ) -> AsyncEventContextManager[FileChooser]:
        return cast(
            AsyncEventContextManager[FileChooser],
            self.page.expect_file_chooser(timeout=timeout),  # type: ignore
        )

    async def check(self, selector: str, *, timeout: float | None = None) -> None:
        await self.page.check(selector, timeout=timeout)  # type: ignore

    async def reload(self, *, timeout: float | None = None) -> None:
        await self.page.reload(timeout=timeout)  # type: ignore


@dataclass(slots=True, frozen=True)
class PlaywrightElementHandle:
    html_element: playwright.ElementHandle

    async def query_selector(self, selector: str) -> Self | None:
        if html_element := await self.html_element.query_selector(selector):
            return PlaywrightElementHandle(html_element)

        return None

    async def query_selector_all(self, selector: str) -> list[Self]:
        return [
            PlaywrightElementHandle(html_element)
            for html_element in await self.html_element.query_selector_all(selector)
        ]

    async def text_content(self) -> str | None:
        return await self.html_element.text_content()

    async def get_attribute(self, name: str) -> str | None:
        return await self.html_element.get_attribute(name)

    async def click(self, *, timeout: float | None = None) -> None:
        await self.html_element.click(timeout=timeout)  # type: ignore

    async def select_option(
        self,
        value: str | list[str] | None = None,
        *,
        label: str | list[str] | None = None,
        timeout: float | None = None,
    ) -> list[str]:
        return await self.html_element.select_option(
            value=value, label=label, timeout=timeout  # type: ignore
        )

    async def scroll_into_view_if_needed(self, *, timeout: float | None = None) -> None:
        await self.html_element.scroll_into_view_if_needed(timeout=timeout)  # type: ignore

    async def focus(self):
        await self.html_element.focus()

    async def is_visible(self) -> bool:
        return await self.html_element.is_visible()

    async def is_enabled(self) -> bool:
        return await self.html_element.is_enabled()

    async def fill(self, value: str, *, timeout: float | None = None) -> None:
        await self.html_element.fill(value, timeout=timeout)  # type: ignore

    async def type(self, text: str, *, timeout: float | None = None) -> None:
        await self.html_element.type(text, timeout=timeout)  # type: ignore

    async def press(self, key: str, *, timeout: float | None = None) -> None:
        await self.html_element.press(key, timeout=timeout)  # type: ignore

    async def inner_html(self) -> str:
        return await self.html_element.inner_html()



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Protocol

from dunia.error import (
    LoginInputNotFound,
    PasswordInputNotFound,
    PlaywrightTimeoutError,
)
from dunia.log import success

if TYPE_CHECKING:
    from dunia.browser import Browser
    from dunia.page import Page


class LoginStrategy(Protocol):
    async def __call__(self, page: Page, login_button_query: str) -> None:
        ...


async def default_login_button_strategy(page: Page, login_button_query: str) -> None:
    # ? Most markets redirect from login page when submitting the form, so we are putting it inside expect_navigation() block
    # ? If it doesn't work for some markets, then remove expect_navigation() and put the code outside it
    async with page.expect_navigation():
        await page.click(login_button_query)

    await page.wait_for_load_state(
        state="load"
    )  # ? "load" is fine for most markets, but some markets don't show full page details until all the network requests are resolved, so for that "networkidle" will be used


@dataclass(slots=True, frozen=True, kw_only=True)
class LoginInfo:
    """
    All the required information that will allow the browser to login to the market website that is going to be crawled. This doesn't contain any information about the type of Browser.
    """

    login_url: str = field(
        metadata={"help": "The url of the login page where input forms are present"}
    )
    user_id: str = field(metadata={"help": "User id or email"})
    password: str = field(metadata={"help": "User password"})
    user_id_query: str = field(
        metadata={"help": "Query for selecting the input field for typing the user id"}
    )
    password_query: str = field(
        metadata={"help": "Query for selecting the input field for typing the password"}
    )
    login_button_query: str = field(
        metadata={
            "help": "Query for selecting the login button or whatever element that is required for submitting the login request"
        }
    )
    keep_logged_in_check_query: str | None = field(
        default=None,
        metadata={
            "help": "Query for the checkbox for keeping the user logged in."
            "Some markets have the checkbox, in such case, it's better to use it to perhaps help the browser store the cache of user credentials"
        },
    )
    login_button_strategy: LoginStrategy = field(
        default=default_login_button_strategy,
        metadata={
            "help": "Most markets provides the clickable submit button, but sometimes the button cannot be clicked."
            'In such case, a different strategy for submitting the credentials will be used (such as pressing "Enter", etc.)'
            'In that case, we can say that it doesn\'t necessarily needs to be login "button" strategy, but anything that can successfully submit the form'
        },
    )


@dataclass(slots=True, frozen=True)
class Login:
    login_info: LoginInfo

    async def __call__(self, browser: Browser) -> None:
        page = await browser.new_page()
        # ? Sometimes if there are network requests happening in the background, the "input" fields keep reverting to the original/previous state while typing
        await page.goto(self.login_info.login_url, wait_until="networkidle")

        is_already_login = True
        try:
            await page.wait_for_selector(
                self.login_info.user_id_query,
                state="visible",
            )

            is_already_login = False
        except PlaywrightTimeoutError:
            is_already_login = True

        if not is_already_login:
            input_id = await page.query_selector(self.login_info.user_id_query)

            if input_id:
                await input_id.fill(self.login_info.user_id)
            else:
                raise LoginInputNotFound(
                    f"User ID ({self.login_info.user_id}) could not be entered"
                )

            if self.login_info.keep_logged_in_check_query:
                await page.check(self.login_info.keep_logged_in_check_query)

            input_password = await page.query_selector(
                self.login_info.password_query,
            )
            if input_password:
                await input_password.fill(self.login_info.password)
            else:
                raise PasswordInputNotFound(
                    f"Passowrd ({self.login_info.password}) could not be entered"
                )

            await self.login_info.login_button_strategy(
                page, self.login_info.login_button_query
            )

            success(
                f"Logged in <MAGENTA><w>(ID: {self.login_info.user_id}, PW: {self.login_info.password})</></>"
            )

        await page.close()



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from functools import cache
from typing import cast

import lxml.html as lxml
from cssselect import HTMLTranslator, SelectorError


@cache
def css_to_xpath(selector: str) -> str | None:
    # ? If the selector is already XPATH, then just return it
    if (
        selector.startswith("xpath=")
        or selector.startswith("//")
        or selector.startswith("..")
    ):
        return selector.removeprefix("xpath=")

    try:
        return HTMLTranslator().css_to_xpath(selector)
    except SelectorError:
        return None


def cssselect(tree: lxml.HtmlElement, selector: str) -> list[lxml.HtmlElement]:
    if expression := css_to_xpath(selector):
        return cast(list[lxml.HtmlElement], tree.xpath(expression))

    return []


def text_content(tree: lxml.HtmlElement, selector: str) -> str | None:
    html_elements = cssselect(tree, selector)
    if len(html_elements):
        return html_elements[0].text_content()

    return None


def inner_text(tree: lxml.HtmlElement, selector: str) -> str | None:
    html_elements = cssselect(tree, selector)
    if len(html_elements):
        return html_elements[0].text

    return None


def get_attribute(tree: lxml.HtmlElement, selector: str, name: str) -> str | None:
    html_elements = cssselect(tree, selector)
    if len(html_elements):
        return html_elements[0].get(name, None)

    return None



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING

from colorama import Fore, init
from playwright.async_api import Error, TimeoutError

from dunia.helpers import compile_regex
from dunia.log import warning

if TYPE_CHECKING:
    from typing import Final


init()


class BasicError(Exception):
    __slots__ = ("message", "url")
    __match_args__: Final = ("message", "url")

    def __init__(self, message: Exception | str, url: str | None = None) -> None:
        self.message = message
        self.url = url

        super().__init__(
            (
                Fore.RED
                + str(self.message)
                + Fore.RESET
                + Fore.CYAN
                + f" || {self.url} ||"
            )
            if self.url
            else (Fore.RED + str(self.message) + Fore.RESET)
        )


class LoginInputNotFound(BasicError):
    pass


class NotAbleToLogin(BasicError):
    pass


class BrowserNotInitialized(BasicError):
    pass


class PasswordInputNotFound(BasicError):
    pass


class TimeoutException(BasicError):
    pass


class HTMLParsingError(BasicError):
    pass


PlaywrightTimeoutError = TimeoutError
PlaywrightError = Error


def backoff_hdlr(details: dict[str, int | float]):

    text = "Backing off {wait:0.1f} seconds after {tries} tries calling function {target} with args {args} and kwargs {kwargs}".format(
        **details
    )

    # ? Fix the loguru's mismatch of <> tag for ANSI color directive
    if source := compile_regex(r"\<\w*\>").findall(text):
        text = text.replace(source[0], source[0].replace("<", r"\<"))

    warning(text)



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
These Protocols can be used in any case of Browser-like class (such as Playwright's BrowserContext, etc.) for static type checking as long as the implementation contains these methods

Sub-classes don't need to explicitly inherit from it to be used in place of this super-class.
"""

from __future__ import annotations

import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, TypedDict, TypeVar

from dunia.page import Page

if TYPE_CHECKING:

    from typing import Final, Literal

    from typing_extensions import Self

    from dunia.request import Request

SCRIPT_PATH: Final[Path] = Path().absolute()
CACHE_DIR: Final[str] = os.path.join(SCRIPT_PATH, "cache")

PageType = TypeVar("PageType", covariant=True)
BrowserContextType = TypeVar("BrowserContextType", covariant=True)


class ViewportSize(TypedDict):
    width: int
    height: int


class ProxySettings(TypedDict, total=False):
    server: str
    bypass: str | None
    username: str | None
    password: str | None


@dataclass(slots=True, frozen=True, kw_only=True)
class BrowserConfig:
    """
    Basic Browser config dataclass that can be used in any kind of browser instance (including Chrome, Firefox, etc.) of any type (i.e., Playwright's Browser or BrowserContext with/without cookies storage state file, etc.) making it browser agnostic.
    """

    user_data_dir: Path | str = field(
        default=CACHE_DIR,
        metadata={"help": "Pass an empty string to use a temporary directory instead"},
    )
    headless: bool = field(
        default=True, metadata={"help": "Whether the browser should be displayed"}
    )
    default_navigation_timeout: int = field(
        default=300000,
        metadata={
            "help": "Time in milliseconds to wait for the new page to load while navigating from the previous page (e.g., it is used in goto(), expect_navigation(), etc.)"
        },
    )
    default_timeout: int = field(
        default=30000,
        metadata={
            "help": "Time in milliseconds to wait for the element to get displayed/enabled that has been queried (e.g., it is used in click(), text_content(), etc.)"
        },
    )
    slow_mo: int | None = field(
        default=None,
        metadata={
            "help": "Delay in milliseconds between actions in the browser (slow motion mode) for debugging purposes. Normally this is not needed as we can simply write unit tests for problematic pages or queries"
        },
    )
    accept_downloads: bool = field(
        default=True,
        metadata={
            "help": "Whether to accept the downloads. This must be set to true in markets where downloading is done by the browser/driver"
        },
    )
    viewport: ViewportSize | None = field(
        default=None,
        metadata={
            "help": "Viewport size. By default it does not enforce fixed viewport, allows resizing window in the headed mode"
        },
    )
    devtools: bool = field(
        default=False,
        metadata={
            "help": "Open the devtools when the browser is launched (it doesn't work on Firefox browser)"
        },
    )
    browser: Literal["chromium", "firefox"] = field(
        default="chromium",
        metadata={"help": "Whether to use Chromium or Firefox browser"},
    )
    channel: Literal[
        "chrome",
        "chrome-beta",
        "chrome-dev",
        "chrome-canary",
        "msedge",
        "msedge-beta",
        "msedge-dev",
        "msedge-canary",
    ] = field(
        default="chrome",
        metadata={"help": "Channel to download browser from"},
    )
    locale: str = field(
        default="ko-KR",
        metadata={"help": "Set language locale"},
    )
    proxy: ProxySettings | None = field(
        default=None,
        metadata={"help": "Proxy server configuration"},
    )


class BrowserPageSpawner(Protocol[PageType]):
    async def new_page(self) -> PageType:
        """This is present here so that new_page() for its sub-classes will work regardless of the type of Browser"""
        ...


class BrowserRequestHandle(Protocol):
    @property
    def request(self) -> Request:
        """Returns the Request object for HTTP methods"""
        ...


class BrowserContext(Protocol[BrowserContextType]):
    @property
    def context(self) -> Any:
        """Current instance of the browser context"""
        ...


class Browser(
    BrowserPageSpawner[Page], BrowserRequestHandle, BrowserContext[Any], Protocol
):
    pass


class BrowserWithLogin(Browser, Protocol):
    async def login(self) -> Self:
        """Returns the self to enable fluent interface programming"""
        ...


class BrowserWithoutLogin(Browser, Protocol):
    async def create(self) -> Self:
        """Returns the self to enable fluent interface programming"""
        ...



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING

from dunia.lxml._core import cssselect, get_attribute, inner_text, text_content

if TYPE_CHECKING:
    import lxml.html as lxml
    from typing_extensions import Self

    from dunia.element import Element


@dataclass(slots=True, frozen=True)
class LXMLDocument:
    html_element: lxml.HtmlElement

    async def query_selector(self, selector: str) -> Element | None:
        html_elements = await asyncio.to_thread(cssselect, self.html_element, selector)
        if len(html_elements):
            return LXMLElement(html_elements[0])

        return None

    async def query_selector_all(self, selector: str) -> list[Element]:
        return [
            LXMLElement(html_element)
            for html_element in await asyncio.to_thread(
                cssselect, self.html_element, selector
            )
        ]

    async def text_content(
        self, selector: str, timeout: int | None = None
    ) -> str | None:
        return await asyncio.to_thread(text_content, self.html_element, selector)

    async def inner_text(self, selector: str, timeout: int | None = None) -> str | None:
        return await asyncio.to_thread(inner_text, self.html_element, selector)

    async def get_attribute(
        self, selector: str, name: str, timeout: int | None = None
    ) -> str | None:
        return await asyncio.to_thread(get_attribute, self.html_element, selector, name)


@dataclass(slots=True, frozen=True)
class LXMLElement:
    html_element: lxml.HtmlElement

    async def query_selector(self, selector: str) -> Self | None:
        html_elements = await asyncio.to_thread(cssselect, self.html_element, selector)
        if len(html_elements):
            return LXMLElement(html_elements[0])

        return None

    async def query_selector_all(self, selector: str) -> list[Self]:
        return [
            LXMLElement(html_element)
            for html_element in await asyncio.to_thread(
                cssselect, self.html_element, selector
            )
        ]

    async def text_content(self) -> str | None:
        if text := await asyncio.to_thread(self.html_element.text_content):
            return text

        return None

    async def get_attribute(self, name: str) -> str | None:
        return self.html_element.get(name, default=None)



from __future__ import annotations

import os
import sys
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from loguru import logger

if TYPE_CHECKING:
    from typing import Final


def get_log_file(logs_dir: str) -> str:
    return os.path.join(logs_dir, f"{datetime.now().strftime('%Y%m%d')}.log")


logger.remove()

LOGGER_FORMAT_STR: Final[
    str
] = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line} [Process ID {extra[id]}]</cyan> - <level>{message}</level>"

logger.add(
    sys.stderr,
    format=LOGGER_FORMAT_STR,
    level="DEBUG",
    colorize=True,
    enqueue=True,
)

PROCESS_ID: Final[int] = os.getpid()

logger = logger.bind(id=PROCESS_ID).opt(colors=True)

logger.add(
    get_log_file(os.path.join(Path().absolute(), "logs")),
    enqueue=True,
    encoding="utf-8-sig",
)

info = logger.info
debug = logger.debug
warning = logger.warning
error = logger.error
success = logger.success



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
import os
from contextlib import suppress
from functools import cache
from typing import TYPE_CHECKING, Any, cast

import backoff
import lxml.html as lxml
from charset_normalizer import detect
from selectolax.lexbor import LexborHTMLParser
from selectolax.parser import HTMLParser
from throttler import throttle

from dunia.aio import with_timeout
from dunia.error import (
    HTMLParsingError,
    PlaywrightError,
    PlaywrightTimeoutError,
    TimeoutException,
    backoff_hdlr,
)
from dunia.lexbor import LexborDocument
from dunia.log import debug
from dunia.lxml import LXMLDocument
from dunia.modest import ModestDocument
from dunia.page import Document

if TYPE_CHECKING:
    from typing import Literal

    from dunia.browser import Browser
    from dunia.html import HTML
    from dunia.page import Page


# ? Sometimes websites are throwing JavaScript exceptions in devtools console, which makes the page stuck on "networkidle", so let's make "load" by default for now
@backoff.on_exception(
    backoff.expo,
    TimeoutException,
    max_tries=5,
    on_backoff=backoff_hdlr,
)
async def visit_link(
    page: Page,
    product_href: str,
    wait_until: Literal["commit", "domcontentloaded", "load", "networkidle"] = "load",
) -> None:
    """
    Visit the page (url) and retry for 5 times if the navigation has been failed within the configured timeout
    """
    try:
        await page.goto(product_href, wait_until=wait_until)
    except (PlaywrightTimeoutError, PlaywrightError) as err:
        raise TimeoutException(err) from err


async def load_html(
    html: HTML,
) -> str | None:
    """
    Utility function to load the html page with "None" type checking/narrowing
    """
    return await html.load() if await html.exists() else None


async def load_content(
    *,
    browser: Browser,
    url: str,
    html: HTML,
    rate_limit: int = 10,
    async_timeout: int = 600,
    save_html: bool = True,
    wait_until: Literal["commit", "domcontentloaded", "load", "networkidle"] = "load",
    strict: bool = False,
) -> str:
    """
    Load HTML content

    Read from the file if it exists on disk, otherwise fetch it with Browser using HTTP's GET request

    If the request fails and 'strict' is False, then visit the URL
    """
    if save_html:
        if await html.exists():
            debug(f"Loading content from existing HTML: {html.file}")
            content = await html.load()
        else:
            try:
                debug(f"Fetching content from URL: {url}")
                content = await fetch_content(browser, url, rate_limit)
            except UnicodeDecodeError as err:
                if strict:
                    raise err from err

                debug(
                    f'Fetching fails due to an error -> "{err}", visiting the URL ({url}) ...'
                )
                visit = with_timeout(async_timeout)(  # type: ignore
                    throttle(rate_limit=rate_limit, period=1.0)(  # type: ignore
                        visit_link
                    )
                )
                page = await browser.new_page()
                await visit(page, url, wait_until=wait_until)
                content = await page.content()
                await page.close()
            await html.save(content)
    else:
        try:
            debug(f"Fetching content from URL: {url}")
            content = await fetch_content(browser, url, rate_limit)
        except UnicodeDecodeError as err:
            if strict:
                raise err from err

            debug(
                f'Fetching fails due to an error -> "{err}", visiting the URL ({url}) ...'
            )
            visit = with_timeout(async_timeout)(  # type: ignore
                throttle(rate_limit=rate_limit, period=1.0)(visit_link)  # type: ignore
            )
            page = await browser.new_page()
            await visit(page, url, wait_until=wait_until)
            content = await page.content()
            await page.close()

    return content


async def reload_content(
    *,
    browser: Browser,
    url: str,
    html: HTML,
    rate_limit: int = 10,
    async_timeout: int = 600,
    save_html: bool = True,
    wait_until: Literal["commit", "domcontentloaded", "load", "networkidle"] = "load",
    strict: bool = False,
) -> str:
    """
    Reload HTML content

    Remove the file if it exists on disk and fetch it again with Browser using HTTP's GET request

    If the request fails and 'strict' is False, then visit the URL
    """
    if save_html:
        with suppress(OSError):
            os.remove(html.file)

    try:
        debug(f"Fetching content from URL: {url}")
        content = await fetch_content(browser, url, rate_limit)
    except UnicodeDecodeError as err:
        if strict:
            raise err from err

        debug(
            f'Fetching fails due to an error -> "{err}", visiting the URL ({url}) ...'
        )
        visit = with_timeout(async_timeout)(  # type: ignore
            throttle(rate_limit=rate_limit, period=1.0)(visit_link)  # type: ignore
        )
        page = await browser.new_page()
        await visit(page, url, wait_until=wait_until)
        content = await page.content()
        await page.close()

    if save_html:
        await html.save(content)

    return content


@backoff.on_exception(
    backoff.expo,
    TimeoutException,
    max_tries=5,
    on_backoff=backoff_hdlr,
)
async def fetch_content(
    browser: Browser, url: str, rate_limit: int, encoding: str | None = None
) -> str:
    """
    Use the Browser to send HTTP's GET request and receive the content response

    If encoding is not provided, then it will try to find the encoding from content body

    If it fails, then encoding will be detected using `charset_normalizer`
    """
    get = throttle(rate_limit=rate_limit, period=1.0)(  # type: ignore
        browser.request.get
    )

    try:
        response: Any = await get(url)
    except (PlaywrightTimeoutError, PlaywrightError) as err:
        raise TimeoutException(err) from err

    body = cast(bytes, await response.body())

    if encoding:
        return body.decode(encoding)

    try:
        content_type = response.headers["content-type"]
    except KeyError:
        detected_encoding = await detect_encoding(body)
        debug(f"Detected encoding: {detected_encoding}")

        return body.decode(detected_encoding)
    else:
        debug(f"Content-Type: {content_type}")

        if "charset=" in content_type:
            content_encoding = content_type.split("charset=")[-1].strip()
            debug(f"Content encoding: {content_encoding}")

            return body.decode(content_encoding)
        else:
            detected_encoding = await detect_encoding(body)
            debug(f"Detected encoding: {detected_encoding}")

            return body.decode(detected_encoding)


@cache
async def detect_encoding(content: bytes) -> str:
    """
    Find the most probable encoding of the content
    """
    encoding = await asyncio.to_thread(detect, content)
    return encoding["encoding"]  # type: ignore


async def load_page(
    *,
    browser: Browser,
    url: str,
    html: HTML,
    rate_limit: int = 10,
    async_timeout: int = 600,
    save_html: bool = True,
    wait_until: Literal["commit", "domcontentloaded", "load", "networkidle"] = "load",
) -> Page:
    """
    Create a new page in the browser and visit the URL

    Save the HTML content of page in the file on disk
    """
    if save_html:
        if await html.exists():
            debug(f"Loading content from existing HTML: {html.file}")
            content = await html.load()
            page = await browser.new_page()
            await page.set_content(content, wait_until=wait_until)
        else:
            debug(f"Visiting the URL ({url}) ...")
            visit = with_timeout(async_timeout)(  # type: ignore
                throttle(rate_limit=rate_limit, period=1.0)(visit_link)  # type: ignore
            )
            page = await browser.new_page()
            await visit(page, url, wait_until=wait_until)
            content = await page.content()
            await html.save(content)
    else:
        debug(f"Visiting the URL ({url}) ...")
        visit = with_timeout(async_timeout)(  # type: ignore
            throttle(rate_limit=rate_limit, period=1.0)(visit_link)  # type: ignore
        )
        page = await browser.new_page()
        await visit(page, url, wait_until=wait_until)

    return page


async def reload_page(
    *,
    browser: Browser,
    url: str,
    html: HTML,
    rate_limit: int = 10,
    async_timeout: int = 600,
    save_html: bool = True,
    wait_until: Literal["commit", "domcontentloaded", "load", "networkidle"] = "load",
) -> Page:
    """
    Create a new page in the browser

    If the file exists on disk, then remove it and visit the URL again
    """
    if save_html:
        with suppress(OSError):
            os.remove(html.file)

    debug(f"Visiting the URL ({url}) ...")
    visit = with_timeout(async_timeout)(  # type: ignore
        throttle(rate_limit=rate_limit, period=1.0)(visit_link)  # type: ignore
    )
    page = await browser.new_page()
    await visit(page, url, wait_until=wait_until)
    content = await page.content()

    if save_html:
        await html.save(content)

    return page


async def parse_document(
    content: str,
    *,
    engine: Literal["lxml", "modest", "lexbor"] = "lxml",
) -> Document | None:
    """
    Parse the HTML content using the specified parser ("lxml", "modest", "lexbor")

    Return document object
    """
    if engine == "lxml":
        try:
            tree = cast(lxml.HtmlElement, await asyncio.to_thread(lxml.fromstring, content))  # type: ignore
        except lxml.etree.ParserError:
            return None

        return LXMLDocument(tree)

    elif engine == "lexbor":
        try:
            tree = await asyncio.to_thread(LexborHTMLParser, content)
        except Exception:
            return None

        return LexborDocument(tree)

    elif engine == "modest":
        try:
            tree = await asyncio.to_thread(HTMLParser, content)
        except Exception:
            return None

        return ModestDocument(tree)

    raise ValueError(
        f'Wrong engine type: {engine}\nSupported engines: ["lxml", "modest", "lexbor"]'
    )


async def parse_document_from_url(
    browser: Browser,
    url: str,
    *,
    rate_limit: int = 10,
    async_timeout: int = 600,
    engine: Literal["lxml", "modest", "lexbor"] = "lxml",
    wait_until: Literal["commit", "domcontentloaded", "load", "networkidle"] = "load",
) -> Document:
    """
    Visit the URL and parse the HTML content using the specified parser ("lxml", "modest", "lexbor")

    Return document object if parsing is successful, however, unlike parse_document(), it raises an HTMLParsingError exception if parsing is failed
    """
    page = await browser.new_page()
    visit = with_timeout(async_timeout)(  # type: ignore
        throttle(rate_limit=rate_limit, period=1.0)(visit_link)  # type: ignore
    )
    await visit(page, url, wait_until=wait_until)
    content = await page.content()
    await page.close()

    if engine == "lxml":
        try:
            tree = cast(
                lxml.HtmlElement,
                await asyncio.to_thread(lxml.fromstring, content),  # type: ignore
            )
        except lxml.etree.ParserError as err:
            raise HTMLParsingError(
                f'Could not parse LXML document due to an error -> "{err}"'
            ) from err

        return LXMLDocument(tree)

    elif engine == "lexbor":
        try:
            tree = await asyncio.to_thread(LexborHTMLParser, content)
        except Exception as err:
            raise HTMLParsingError(
                f'Could not parse LEXXBOR document due to an error -> "{err}"'
            ) from err

        return LexborDocument(tree)

    elif engine == "modest":
        try:
            tree = await asyncio.to_thread(HTMLParser, content)
        except Exception as err:
            raise HTMLParsingError(
                f'Could not parse MODEST document due to an error -> "{err}"'
            ) from err

        return ModestDocument(tree)

    raise ValueError(
        f'Wrong engine type: {engine}\nSupported engines: ["lxml", "modest", "lexbor"]'
    )



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
from functools import wraps
from typing import TYPE_CHECKING

from async_timeout import timeout as async_timeout

if TYPE_CHECKING:
    from asyncio import Future
    from collections.abc import Awaitable, Callable, Coroutine, Generator
    from typing import Any, Literal, ParamSpec, TypeVar, overload

    ParamsType = ParamSpec("ParamsType")
    ReturnType = TypeVar("ReturnType")

    _T = TypeVar("_T")
    _T1 = TypeVar("_T1")
    _T2 = TypeVar("_T2")
    _T3 = TypeVar("_T3")
    _T4 = TypeVar("_T4")
    _T5 = TypeVar("_T5")
    _T6 = TypeVar("_T6")
    _T7 = TypeVar("_T7")
    _T8 = TypeVar("_T8")
    _FutureT = Future[_T] | Generator[Any, None, _T] | Awaitable[_T]

    # ? Typing for asyncio.gather() for upto 8 coroutines
    # ? Modified from Playwright's typeshed
    @overload
    def gather(  # type: ignore
        __coro_or_future1: _FutureT[_T1], *, return_exceptions: Literal[False] = ...
    ) -> Future[tuple[_T1]]:
        ...

    @overload
    def gather(  # type: ignore
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        *,
        return_exceptions: Literal[False] = ...,
    ) -> Future[tuple[_T1, _T2]]:
        ...

    @overload
    def gather(  # type: ignore
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        *,
        return_exceptions: Literal[False] = ...,
    ) -> Future[tuple[_T1, _T2, _T3]]:
        ...

    @overload
    def gather(  # type: ignore
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        __coro_or_future4: _FutureT[_T4],
        *,
        return_exceptions: Literal[False] = ...,
    ) -> Future[tuple[_T1, _T2, _T3, _T4]]:
        ...

    @overload
    def gather(  # type: ignore
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        __coro_or_future4: _FutureT[_T4],
        __coro_or_future5: _FutureT[_T5],
        *,
        return_exceptions: Literal[False] = ...,
    ) -> Future[tuple[_T1, _T2, _T3, _T4, _T5]]:
        ...

    @overload
    def gather(  # type: ignore
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        __coro_or_future4: _FutureT[_T4],
        __coro_or_future5: _FutureT[_T5],
        __coro_or_future6: _FutureT[_T6],
        *,
        return_exceptions: Literal[False] = ...,
    ) -> Future[tuple[_T1, _T2, _T3, _T4, _T5, _T6]]:
        ...

    @overload
    def gather(  # type: ignore
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        __coro_or_future4: _FutureT[_T4],
        __coro_or_future5: _FutureT[_T5],
        __coro_or_future6: _FutureT[_T6],
        __coro_or_future7: _FutureT[_T7],
        *,
        return_exceptions: Literal[False] = ...,
    ) -> Future[tuple[_T1, _T2, _T3, _T4, _T5, _T6, _T7]]:
        ...

    @overload
    def gather(  # type: ignore
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        __coro_or_future4: _FutureT[_T4],
        __coro_or_future5: _FutureT[_T5],
        __coro_or_future6: _FutureT[_T6],
        __coro_or_future7: _FutureT[_T7],
        __coro_or_future8: _FutureT[_T8],
        *,
        return_exceptions: Literal[False] = ...,
    ) -> Future[tuple[_T1, _T2, _T3, _T4, _T5, _T6, _T7, _T8]]:
        ...

    @overload
    def gather(
        __coro_or_future1: _FutureT[Any],
        __coro_or_future2: _FutureT[Any],
        __coro_or_future3: _FutureT[Any],
        __coro_or_future4: _FutureT[Any],
        __coro_or_future5: _FutureT[Any],
        __coro_or_future6: _FutureT[Any],
        __coro_or_future7: _FutureT[Any],
        __coro_or_future8: _FutureT[Any],
        __coro_or_future9: _FutureT[Any],
        __coro_or_future10: _FutureT[Any],
        *coros_or_futures: _FutureT[Any],
        return_exceptions: bool = ...,
    ) -> Future[list[Any]]:
        ...

    @overload
    def gather(
        __coro_or_future1: _FutureT[_T1], *, return_exceptions: bool = ...
    ) -> Future[tuple[_T1 | BaseException]]:
        ...

    @overload
    def gather(
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        *,
        return_exceptions: bool = ...,
    ) -> Future[tuple[_T1 | BaseException, _T2 | BaseException]]:
        ...

    @overload
    def gather(
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        *,
        return_exceptions: bool = ...,
    ) -> Future[tuple[_T1 | BaseException, _T2 | BaseException, _T3 | BaseException]]:
        ...

    @overload
    def gather(
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        __coro_or_future4: _FutureT[_T4],
        *,
        return_exceptions: bool = ...,
    ) -> Future[
        tuple[
            _T1 | BaseException,
            _T2 | BaseException,
            _T3 | BaseException,
            _T4 | BaseException,
        ]
    ]:
        ...

    @overload
    def gather(
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        __coro_or_future4: _FutureT[_T4],
        __coro_or_future5: _FutureT[_T5],
        *,
        return_exceptions: bool = ...,
    ) -> Future[
        tuple[
            _T1 | BaseException,
            _T2 | BaseException,
            _T3 | BaseException,
            _T4 | BaseException,
            _T5 | BaseException,
        ]
    ]:
        ...

    @overload
    def gather(
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        __coro_or_future4: _FutureT[_T4],
        __coro_or_future5: _FutureT[_T5],
        __coro_or_future6: _FutureT[_T6],
        *,
        return_exceptions: bool = ...,
    ) -> Future[
        tuple[
            _T1 | BaseException,
            _T2 | BaseException,
            _T3 | BaseException,
            _T4 | BaseException,
            _T5 | BaseException,
            _T6 | BaseException,
        ]
    ]:
        ...

    @overload
    def gather(
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        __coro_or_future4: _FutureT[_T4],
        __coro_or_future5: _FutureT[_T5],
        __coro_or_future6: _FutureT[_T6],
        __coro_or_future7: _FutureT[_T7],
        *,
        return_exceptions: bool = ...,
    ) -> Future[
        tuple[
            _T1 | BaseException,
            _T2 | BaseException,
            _T3 | BaseException,
            _T4 | BaseException,
            _T5 | BaseException,
            _T6 | BaseException,
            _T7 | BaseException,
        ]
    ]:
        ...

    @overload
    def gather(
        __coro_or_future1: _FutureT[_T1],
        __coro_or_future2: _FutureT[_T2],
        __coro_or_future3: _FutureT[_T3],
        __coro_or_future4: _FutureT[_T4],
        __coro_or_future5: _FutureT[_T5],
        __coro_or_future6: _FutureT[_T6],
        __coro_or_future7: _FutureT[_T7],
        __coro_or_future8: _FutureT[_T8],
        *,
        return_exceptions: bool = ...,
    ) -> Future[
        tuple[
            _T1 | BaseException,
            _T2 | BaseException,
            _T3 | BaseException,
            _T4 | BaseException,
            _T5 | BaseException,
            _T6 | BaseException,
            _T7 | BaseException,
            _T8 | BaseException,
        ]
    ]:
        ...


async def gather(  # type: ignore
    *coros_or_futures: _FutureT[Any],
    return_exceptions: bool = False,
) -> tuple[Any | BaseException, ...]:
    return await asyncio.gather(*coros_or_futures, return_exceptions=return_exceptions)  # type: ignore


def with_timeout(timeout: int):
    def decorator(
        fn: Callable[ParamsType, Coroutine[Any, Any, ReturnType]]
    ) -> Callable[ParamsType, Coroutine[Any, Any, ReturnType]]:
        @wraps(fn)
        async def wrapper(*args: ParamsType.args, **kwargs: ParamsType.kwargs):
            async with async_timeout(timeout):
                return await fn(*args, **kwargs)

        return wrapper

    return decorator



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from typing_extensions import Self


class StateSaver(Protocol):
    async def save(self) -> None:
        """Save the state class object to pickle file (for tracking which pages have been crawled)."""
        ...


class StateLoader(Protocol):
    async def load(self) -> Self:
        """Load the state class object from pickle file."""
        ...


class StateFile(Protocol):
    @property
    def directory(self) -> str:
        """Full directory for saved pickle file."""
        ...

    @property
    def file(self) -> str:
        """Full path for saved pickle file."""
        ...

    async def exists(self) -> bool:
        """Whether the pickle file exists."""
        ...


class State(StateSaver, StateLoader, StateFile, Protocol):
    pass



# MIT License

# Copyright (c) 2022 Danyal Zia Khan

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from selectolax.lexbor import LexborHTMLParser, LexborNode


async def css_first(document_or_node: LexborHTMLParser | LexborNode, selector: str):
    if (splitter := ",") in selector or (splitter := ", ") in selector:
        for selector in selector.split(splitter):
            if html_element := await asyncio.to_thread(
                document_or_node.css_first, selector, default=None
            ):
                return html_element
    elif html_element := await asyncio.to_thread(
        document_or_node.css_first, selector, default=None
    ):
        return html_element


async def css(document_or_node: LexborHTMLParser | LexborNode, selector: str):
    if (splitter := ",") in selector or (splitter := ", ") in selector:
        tasks = (
            asyncio.to_thread(document_or_node.css, selector)
            for selector in selector.split(splitter)
        )
        all_css = await asyncio.gather(*tasks)
        return [
            html_element for html_elements in all_css for html_element in html_elements
        ]

    return await asyncio.to_thread(document_or_node.css, selector)
